import csv

from schwifty import BIC, IBAN, exceptions

from ._builtin import Page, WaitPage
from .encryption import encrypt_payment_data
from .models import output_col_names

replacements = {
    "ä": "ae",
    "ö": "oe",
    "ü": "ue",
    "ß": "ss",
    "Ä": "Ae",
    "Ö": "Oe",
    "Ü": "Ue",
}


def clean_up(payment_data: dict):
    """Remove obvious mistakes from the payment data.

    Only mistakes where the replacement is clear and it does not need to be send back to the user."""
    for key in payment_data:
        payment_data[key] = (
            payment_data[key].replace(",", " ").replace("  ", " ").strip()
        )
    payment_data["iban"] = (
        payment_data["iban"].upper().replace(" ", "").replace(".", "")
    )
    payment_data["bic"] = payment_data["bic"].upper().replace(" ", "").replace(".", "")
    for key in replacements.keys():
        payment_data["full_name"] = payment_data["full_name"].replace(
            key, replacements[key]
        )
    return payment_data


def sort_and_clean_data(post_data):
    payment_data = {col: post_data[col] for col in output_col_names}
    payment_data = clean_up(payment_data)
    return payment_data


class WaitingForOthers(WaitPage):
    def is_displayed(self):
        return not self.session.config["disable_waiting_for_others"]


class PaymentInfos(Page):
    def vars_for_template(self):
        return {
            "final_payoff": self.participant.payoff_plus_participation_fee(),
            "saved": self.participant.vars["temp_values"],
        }

    def error_message(self, value):
        payment_data = sort_and_clean_data(self.request.POST)
        self.participant.vars[
            "temp_values"
        ] = payment_data  # temp saves them, removed in before_next_page

        name = payment_data["full_name"].split()
        if len(name) < 2:
            return "Please enter first and last name."
        for name_part in name:
            if not name_part.isascii() or not name_part.isalpha():
                return "Please only use the standard latin alphabet."

        try:
            iban = IBAN(payment_data["iban"])
        except exceptions.SchwiftyException as error:
            return error
        if iban.country_code not in ["AT", "DE"]:
            try:
                BIC(payment_data["bic"])
            except exceptions.SchwiftyException as error:
                return error

    def before_next_page(self):
        payment_data = sort_and_clean_data(self.request.POST)
        self.participant.vars[
            "temp_values"
        ].clear()  # removes temp saved iban/name, etc
        # transforming to int as is otherwise in c(), which might cause problems when encrypting/saving
        # amount * 100 to prevent floating problems
        payment_data["amount"] = int(
            self.participant.payoff_plus_participation_fee() * 100
        )
        if self.session.config["encrypt_payment_file"]:
            payment_data = encrypt_payment_data(payment_data)
        with open("payment_info.csv", "a") as output_file:
            csv.DictWriter(output_file, fieldnames=payment_data.keys()).writerow(
                payment_data
            )
            output_file.close()


class ConfirmationPage(Page):
    pass


page_sequence = [WaitingForOthers, PaymentInfos, ConfirmationPage]



from itertools import product

from otree.api import Submission, SubmissionMustFail

from . import pages
from ._builtin import Bot

correct_names = [
    "Detlef Mueller",
    "Detlef Müller",
    "Detlef Hans Dieter Müller",
    " Detlef Hans Dieter Müller  ",
]


incorrect_names = ["ichnureinennamenhaben", "Chloé IsNotAllowed", "陈玉"]

# Combination iban + bic
correct_iban_bic = [
    ["DE40100100100000012345", ""],
    ["DE75512108001245126199", ""],
    ["de75512108001245126199", ""],
    ["de 75 512 108 001 245  126199", ""],
    ["DE75512108001245126199", "-"],
    ["DE75512108001245126199", "nix"],
    ["GR1601101250000000012300695", "ETHNGRAA"],
    ["GB33BUKB20201555555555", "BUKBGB22"],
    ["FR7630006000011234567890189", "AGRIFRPP"],
    ["AT611904300234573201", "GIBAATWW"],
    ["AT611904300234573201", ""],
]

incorrect_iban_bic = [
    ["DE40100100100000012346", ""],
    ["GR1601101250000000012300695", ""],
    ["GB33BUKB20201555555555", ""],
    ["FR7630006000011234567890189", ""],
    ["AT6119043002345732012", "GIBAATWW"],
    ["AT61190430023453201", "GIBAATWW"],
]

all_good_cases = product(correct_names, correct_iban_bic)
all_bad_cases_1 = product(correct_names, incorrect_iban_bic)
all_bad_cases_2 = product(incorrect_names, correct_iban_bic)
all_very_bad_cases = product(incorrect_names, incorrect_iban_bic)


all_good_cases_dict = [
    {"full_name": entry[0], "iban": entry[1][0], "bic": entry[1][1], "is_correct": True}
    for entry in all_good_cases
]

all_bad_cases_dict_1 = [
    {
        "full_name": entry[0],
        "iban": entry[1][0],
        "bic": entry[1][1],
        "is_correct": False,
    }
    for entry in all_bad_cases_1
]

all_bad_cases_dict_2 = [
    {
        "full_name": entry[0],
        "iban": entry[1][0],
        "bic": entry[1][1],
        "is_correct": False,
    }
    for entry in all_bad_cases_2
]

all_very_bad_cases_dict = [
    {
        "full_name": entry[0],
        "iban": entry[1][0],
        "bic": entry[1][1],
        "is_correct": False,
    }
    for entry in all_very_bad_cases
]

all_cases = (
    all_good_cases_dict
    + all_bad_cases_dict_1
    + all_bad_cases_dict_2
    + all_very_bad_cases_dict
)


class PlayerBot(Bot):

    cases = all_cases

    def play_round(self):
        if self.case["is_correct"]:
            yield Submission(pages.PaymentInfos, self.case, check_html=False)
        else:
            yield SubmissionMustFail(pages.PaymentInfos, self.case, check_html=False)



import base64

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa

# Generates a private/public key pair, using rsa with key size 4096
# key_size of 4096 is such an overkill
# But given that we are not encrypting much, it shouldn't matter
# public_exponent: commonly used one,


def generate_key_pair():
    """ Create private/public key pair and saves them. """
    private_key = rsa.generate_private_key(
        public_exponent=65537, key_size=4096, backend=default_backend()
    )
    public_key = private_key.public_key()

    # Saving the private key
    serial_private = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    with open("../private_noshare.pem", "wb") as f:
        f.write(serial_private)

    # Saving the public key
    serial_pub = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    with open("../public_shared.pem", "wb") as f:
        f.write(serial_pub)


def read_public_key(filename="public_shared.pem"):
    with open("public_shared.pem", "rb") as key_file:
        public_key = serialization.load_pem_public_key(
            key_file.read(), backend=default_backend()
        )
    return public_key


def encrypt_payment_data(payment_data: dict):
    """Encrypt the payment data.
    dict of payment data -> dict of encrypted payment data"""
    public_key = read_public_key()
    for key in payment_data.keys():
        entry = bytes(str(payment_data[key]), "utf-8")
        encrypted_entry = public_key.encrypt(
            entry,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )
        payment_data[key] = base64.standard_b64encode(encrypted_entry)
    return payment_data



import os

from otree.api import BaseConstants, BaseGroup, BasePlayer, BaseSubsession

author = "Philipp Külpmann"

doc = """
App to gather payment data from subjects
"""

output_col_names = ["full_name", "iban", "bic"]


class Constants(BaseConstants):
    name_in_url = "vceepayment"
    players_per_group = None
    num_rounds = 1


class Subsession(BaseSubsession):
    def creating_session(self):
        if (
            not os.path.isfile("public_shared.pem")
            and self.session.config["encrypt_payment_file"]
        ):
            raise FileNotFoundError(
                "Encryption enabled but no public key found! Either upload the public key or disable encryption."
            )

        for player in self.get_players():
            player.participant.vars["temp_values"] = {
                item: "" for item in output_col_names
            }


class Group(BaseGroup):
    pass


class Player(BasePlayer):
    pass
